import io
import logging
import ctypes as c
from collections import deque, defaultdict

from edosl0util.headers import (
    PrimaryHeader,
    GROUP_FIRST,
    GROUP_CONTINUING,
    GROUP_LAST,
    GROUP_STANDALONE
)

from edosl0util import headers

LOG = logging.getLogger(__name__)

# Max CCSDS sequence id
MAX_SEQ_ID = 2**14 - 1


def simple_stream(fobj):
    """
    Generator that yields PrimaryHeaders and data. Files are read using mmap.
    """
    psize = c.sizeof(PrimaryHeader)
    while True:
        buf = fobj.read(psize)
        if len(buf) < psize:
            return
        h1 = PrimaryHeader.from_buffer_copy(buf)
        # read user data
        size = h1.data_length_minus1 + 1
        buf = fobj.read(size)
        if len(buf) < size:
            return
        yield h1, buf


def _full_stream(header_lookup, fobj):
    """
    Generator that yields primary and secondary header and any left over bytes
    of "user" data.

    :param header_lookup: function that takes a primary header and returns
        a secondary header struct to use, or None.
    """
    stream = simple_stream(fobj)
    for h1, buf in stream:
        h2 = ''
        if h1.secondary_header_flag:
            H2Impl = header_lookup(h1)
            if H2Impl:
                hsize = c.sizeof(H2Impl)
                h2 = H2Impl.from_buffer_copy(buf)
                yield h1, h2, buf[hsize:]
                continue
        yield h1, h2, buf


def jpss_full_stream(fobj):
    """
    `_full_stream` impl for JPSS secondary headers.
    """
    return _full_stream(headers.jpss_header_lookup, fobj)


def aqua_full_stream(fobj):
    """
    `_full_stream` impl for Aqua secondary headers.
    """
    return _full_stream(headers.aqua_header_lookup, fobj)


class Packet(object):
    def __init__(self, h1, h2, data):
        self.data = data
        self.primary_header = h1
        self.secondary_header = h2

    def __str__(self):
        return '<Packet apid=%d seqid=%d stamp=%s>' % \
               (self.apid, self.seqid, self.stamp)
    __repr__ = __str__

    @property
    def apid(self):
        return self.primary_header.apid

    @property
    def seqid(self):
        return self.primary_header.source_sequence_count

    @property
    def stamp(self):
        h2 = self.secondary_header
        return h2.timecode.asdatetime() if hasattr(h2, 'timecode') else None

    def bytes(self):
        return buffer(self.primary_header) + \
            buffer(self.secondary_header) + self.data

    def is_group(self):
        return False

    def is_first(self):
        return self.primary_header.sequence_grouping == GROUP_FIRST

    def is_continuine(self):
        return self.primary_header.sequence_grouping == GROUP_CONTINUING

    def is_last(self):
        return self.primary_header.sequence_grouping == GROUP_LAST

    def is_standalone(self):
        return self.primary_header.sequence_grouping == GROUP_STANDALONE


class PacketStream(object):
    """
    Iterates over all CCSDS data producing `Packet` tuples. Only as much data is
    necessary to generate a single packet is read into memory at a time.
    """
    def __init__(self, fobj):
        """
        :param fobj: File-like object to read stream from
        :keyword fail_on_missing: If True raise a MissingPackets error when packets
            are missing.
        """
        self._stream = jpss_full_stream(fobj)
        self._seek_cache = deque()
        self._apid_info = defaultdict(lambda: {'count': 0, 'last_seqid': -1, 'missing': 0})

    def __iter__(self):
        return self

    # python3
    def __next__(self):
        return self.next()

    def push_back(self, packet):
        self._seek_cache.append(packet)

    def next(self):
        """
        Return the next packet in the stream.

        :keyword fail_on_missing: If True MissingPackets will be raised when
            sequence ids are non-sequential.

        :raises: StopIteration when no more data can be serialzed into Packets.
        :raises: MissingPackets see `fail_on_missing`
        """
        # return any items on the seek cache before reading more
        if len(self._seek_cache):
            return self._seek_cache.popleft()

        packet = Packet(*self._stream.next())
        self._update_info(packet)
        return packet

    def _update_info(self, packet):
        apid = self._apid_info[packet.apid]
        apid['count'] += 1
        if apid['last_seqid'] != -1:
            if packet.seqid > apid['last_seqid']:
                missing = packet.seqid - apid['last_seqid'] - 1
            else:
                missing = packet.seqid - apid['last_seqid'] + MAX_SEQ_ID
            apid['missing'] += missing
        apid['last_seqid'] = packet.seqid

    def info(self):
        return self._apid_info

    def seek(self, stamp, apid=None):
        """
        Seek forward such that the next packet provided will be the first packet
        having a timestamp available in the secondary header >= stamp, and
        optionally with apid as well.

        :raises: StopIteration if the end of the stream is reached before the
                 criteria can be met.
        """
        # seek past partial groups
        packet = self.next()
        while not packet.stamp:
            packet = self.next()

        # seek to first packet > `stamp`
        current_stamp = packet.stamp
        while current_stamp <= stamp:
            packet = self.next()
            current_stamp = packet.stamp or current_stamp

        # put back so next packet is the first packet > `stamp`
        self.push_back(packet)

        if apid is not None:
            packet = self.next()
            while packet.apid != apid:
                packet = self.next()
            self.push_back(packet)


def _cmp_by_first_pkt(filea, fileb):
    """
    Compare 2 L0 files by their first available timestamp. Used to sort L0 files
    by time.
    """
    stampa = stampb = None
    for pkt in PacketStream(io.open(filea, 'rb')):
        if pkt.stamp:
            stampa = pkt.stamp
            break
    for pkt in PacketStream(io.open(fileb, 'rb')):
        if pkt.stamp:
            stampb = pkt.stamp
            break
    return cmp(stampa, stampb)


def make_streams(filepaths):
    filepaths.sort(_cmp_by_first_pkt)
    return [PacketStream(io.open(f, 'rb')) for f in filepaths]
